from yta_general_utils.file.checker import file_is_image_file
from yta_general_utils.file.filename import filename_is_type
from yta_general_utils.file.enums import FileType
from yta_general_utils.image.converter import numpy_image_to_pil, pil_image_to_numpy
from PIL import Image
from typing import Union

import numpy as np


class ImageParser:
    """
    Class to simplify the way we handle the image parameters
    so we can parse them as Pillow images, as numpy arrays,
    etc.
    """
    @classmethod
    def to_pillow(cls, image: Union[str, Image.Image, np.ndarray], mode: str = 'RGB'):
        """
        Returns an instance of a Pillow Image.Image of the given
        'image' if it is a valid image and no error found.
        """
        # TODO: By now we are only accepting string filenames,
        # Pillow Image.Image instances and numpy arrays.
        if image is None:
            raise Exception('No "image" parameter provided.')
        
        if not isinstance(image, (str, Image.Image, np.ndarray)):
            raise Exception('The "image" parameter provided is not a string nor a Image.Image nor a np.ndarray.')

        if not mode:
            mode = 'RGB'

        if mode not in ['RGB', 'RGBA']:
            raise Exception('The provided "mode" parameters is not a valid mode: RGB or RGBA.')

        # We can have problems with np.ndarray
        if isinstance(image, np.ndarray):
            image = numpy_image_to_pil(image)
        elif isinstance(image, str):
            if not filename_is_type(image, FileType.IMAGE):
                raise Exception('The "image" parameter provided is not a valid image filename.')
            
            if not file_is_image_file(image):
                raise Exception('The "image" parameter provided is not a valid image.')
            
            image = Image.open(image)

        return image.convert(mode)
    
    @classmethod
    def to_numpy(cls, image: Union[str, Image.Image, np.ndarray], mode: str = 'RGB'):
        """
        Returns a numpy array representing the given 'image'
        if it is a valid image and no error found.

        The 'mode' parameter will be used to open the image
        with Pillow library and then turning into numpy when
        necessary. Must be 'RGB' or 'RGBA'.
        """
        # TODO: By now we are only accepting string filenames,
        # Pillow Image.Image instances and numpy arrays.
        if image is None:
            raise Exception('No "image" parameter provided.')
        
        if not isinstance(image, (str, Image.Image, np.ndarray)):
            raise Exception('The "image" parameter provided is not a string nor a Image.Image nor a np.ndarray.')
        
        if not mode:
            mode = 'RGB'

        if mode not in ['RGB', 'RGBA']:
            raise Exception('The provided "mode" parameters is not a valid mode: RGB or RGBA.')

        # We can have problems with np.ndarray
        if isinstance(image, Image.Image):
            image = pil_image_to_numpy(image.convert(mode))
        elif isinstance(image, str):
            if not filename_is_type(image, FileType.IMAGE):
                raise Exception('The "image" parameter provided is not a valid image filename.')
            
            if not file_is_image_file(image):
                raise Exception('The "image" parameter provided is not a valid image.')
            
            image = pil_image_to_numpy(Image.open(image).convert(mode))

        return image


from yta_general_utils.programming.error_message import ErrorMessage
from yta_general_utils.file.filename import filename_is_type
from yta_general_utils.checker.type import variable_is_positive_number
from yta_general_utils.file.checker import file_exists
from yta_general_utils.checker.url import url_is_ok
from enum import Enum
from typing import Union


class ParameterValidator:
    @classmethod
    def validate_mandatory_parameter(cls, name: str, value):
        """
        Validates if the provided 'value' parameter with the also
        provided 'name' has a value, raising an Exception if not.

        This method returns the provided 'value' if everything is
        ok.
        """
        if not value:
            raise Exception(ErrorMessage.parameter_not_provided(name))

        return value
        
    @classmethod
    def validate_string_parameter(cls, name: str, value: str):
        """
        Validates if the provided 'value' parameter with the also
        provided 'name' is a string value, raising an Exception if
        not.

        This method returns the provided 'value' if everything is
        ok.
        """
        if not isinstance(value, str):
            raise Exception(ErrorMessage.parameter_is_not_string(name))

        return value

    @classmethod
    def validate_bool_parameter(cls, name: str, value: bool):
        """
        Validates if the provided 'value' parameter with the also
        provided 'name' is a boolean value, raising and Exception 
        if not.

        This method returns the provided 'value' if everything is
        ok.
        """
        if not isinstance(value, bool):
            raise Exception(ErrorMessage.parameter_is_not_boolean(name))

        return value
        
    @classmethod
    def validate_file_exists(cls, name: str, value: str):
        """
        Validates if the provided 'value' parameter with the also
        provided 'name' is a file that actually exists, raising
        an Exception if not.

        This method returns the provided 'value' if everything is
        ok.
        """
        if not file_exists(value):
            raise Exception(ErrorMessage.parameter_is_file_that_doesnt_exist(name))

        return value
        
    @classmethod
    def validate_filename_is_type(cls, name: str, value: str, file_type: 'FileType'):
        """
        Validates if the provided 'value' parameter with the also
        provided 'name' is a filename of the given 'file_type',
        raising an Exception if not.

        This method returns the provided 'value' if everything is
        ok.
        """
        if not filename_is_type(value, file_type):
            raise Exception(ErrorMessage.parameter_is_not_file_of_file_type(name, file_type))

        return value
        
    @classmethod
    def validate_url_is_ok(cls, name: str, value: str):
        """
        Validates if the provided 'value' parameter with the also
        provided 'name' is a valid url (the url is accessible),
        raising an Exception if not.

        This method returns the provided 'value' if everything is
        ok.
        """
        if not url_is_ok(value):
            raise Exception(ErrorMessage.parameter_is_not_valid_url(name))

        return value
        
    @classmethod
    def validate_positive_number(cls, name: str, value: Union[int, float]):
        """
        Validates if the provided 'value' parameter with the also
        provided 'name' is a positive number (0 or greater),
        raising an Exception if not.

        This method returns the provided 'value' as it is if 
        everything is ok.
        """
        if not variable_is_positive_number(value):
            raise Exception(ErrorMessage.parameter_is_not_positive_number(name))

        return value

    @classmethod
    def validate_is_class(cls, name: str, value, class_names: Union[list[str], str]):
        """
        Validates if the provided 'value' is one of the provided 'class_names'
        by using the 'type(value).__name__' checking.

        This method returns the 'value' as it is if everything is ok.
        """
        if isinstance(class_names, str):
            class_names = [class_names]

        if not type(value).__name__ in class_names:
            raise Exception(ErrorMessage.parameter_is_not_class(name, class_names))
        
        return value
        
    # Complex ones below
    @classmethod
    def validate_string_mandatory_parameter(cls, name: str, value: str):
        """
        Validates if the provided 'value' is a valid and non
        empty string.
        """
        cls.validate_mandatory_parameter(name, value)
        cls.validate_string_parameter(name, value)

        return value

    @classmethod
    def validate_numeric_positive_mandatory_parameter(cls, name: str, value: str):
        """
        Validates if the provided 'value' is a positive numeric
        value.
        """
        cls.validate_mandatory_parameter(name, value)
        cls.validate_positive_number(name, value)

        return value
    
    @classmethod
    def validate_is_enum_class(cls, enum: Union['YTAEnum', Enum]):
        """
        Validates if the provided 'value' is a valid Enum
        class or subclass.

        This method will raise an Exception if the provided
        'value' is not a valid Enum class or subclass, or
        will return it as it is if yes.
        """
        if not isinstance(enum, Enum) and not issubclass(enum, Enum):
            raise Exception(f'The parameter "{enum}" provided is not an Enum class or subclass.')
        
        return enum

    @classmethod
    def validate_enum(cls, value: Union['YTAEnum', str], enum: 'YTAEnum'):
        """
        Validates if the provided 'value' value is a valid
        Enum or Enum value of the also provided 'enum' class.

        This method will raise an Exception if something is
        wrong or will return the 'value' as an 'enum' Enum.
        instance if everything is ok.
        """
        cls.validate_mandatory_parameter('value', value)
        cls.validate_is_enum_class(enum)
        cls.validate_is_class('value', value, [enum.__class__.__name__, 'str'])

        return enum.to_enum(value)


from yta_general_utils.file.checker import file_exists

import json


def read_json_from_file(filename: str):
    """
    Reads the provided 'filename' and turns the information 
    into a json format (if possible). This method returns
    None if it was not possible.

    @param
        **filename**
        File path from which we want to read the information.
    """
    if not filename:
        return None
    
    if not file_exists(filename):
        return None
    
    with open(filename, encoding = 'utf-8') as json_file:
        return json.load(json_file)


#from yta_general_utils.programming.enum import YTAEnum as Enum
from enum import Enum


class FileSearchOption(Enum):
    """
    Enum that allows us setting the strategy dynamically when 
    searching for files.
    """
    FILES_AND_FOLDERS = 'fifo'
    """
    This option, when set, will return files and folders.
    """
    FILES_ONLY = 'fi'
    """
    This option, when set, will return files only.
    """
    FOLDERS_ONLY = 'fo'
    """
    This option, when set, will return folders only.
    """

class FileType(Enum):
    """
    Enum that represents the different file types and the valid
    extensions we accept for those file types. This Enum is to
    be used when checking filenames parameter.

    For example, we will use this to make sure the filename they
    gave to us is a video file type if we are storing a video 
    file.
    """
    IMAGE = ['.png', '.jpeg', '.jpg', '.webp', '.bmp']
    AUDIO = ['.wav', '.mp3', '.m4a'] # TODO: Add more types
    VIDEO = ['.mov', '.mp4', '.webm', '.avi'] # TODO: Add more types


# TODO: Find a more specific and better way of classifying these
# files and methods. I moved here to avoid using __init__.py

from yta_general_utils.file.enums import FileSearchOption
from yta_general_utils.file.filename import sanitize_filename
from yta_general_utils.file.checker import is_file, is_folder
from shutil import move as shutil_move, copyfile as shutil_copyfile

import glob


def list(abspath, option: FileSearchOption = FileSearchOption.FILES_AND_FOLDERS, pattern: str = '*', recursive: bool = False):
    """
    List what is inside the provided 'abspath'. This method will list files and
    folders, files or only folders attending to the provided 'option'. It will
    also filter the files/folders that fit the provided 'pattern' (you can use
    '*' as wildcard, so for example '*.jpg' will list all images). This method
    can also be used in a recursive way if 'recursive' parameter is True, but
    take care of memory consumption and it would take its time to perform.

    This method returns a list with all existing elements absolute paths 
    sanitized.
    """
    if not abspath:
        return None
    
    abspath = sanitize_filename(abspath)
    list = []

    # This below get files and folders
    files_and_folders = [sanitize_filename(f) for f in glob.glob(pathname = abspath + pattern, recursive = recursive)]

    if option == FileSearchOption.FILES_ONLY:
        for f in files_and_folders:
            if is_file(f):
                list.append(f)
    elif option == FileSearchOption.FOLDERS_ONLY:
        for f in files_and_folders:
            if is_folder(f):
                list.append(f)
    elif option == FileSearchOption.FILES_AND_FOLDERS:
        list = files_and_folders
    
    return list

def rename_file(origin_filename: str, destination_filename: str, replace_if_existing: bool = False):
    """
    Renames the 'origin_filename' to the 'destination_filename'.
    If 'replace_if_existing' is True, it will replace the destination
    file if possible and allowed. If it is False, it will fail.

    TODO: Remove 'replace_if_existing' if not used.
    """
    if not origin_filename:
        return None
    
    if not destination_filename:
        return None
    
    shutil_move(origin_filename, destination_filename)

def copy_file(origin_filename: str, destination_filename: str):
    """
    Makes a copy of the provided 'origin_filename' and 
    stores it as 'destination_filename'.

    The destination folder must exist.
    """
    if not origin_filename:
        return None

    if not destination_filename:
        return None

    shutil_copyfile(origin_filename, destination_filename)


from yta_general_utils.file.checker import is_file


def delete_file(filename: str):
    """
    Deletes the provided 'filename' if existing.

    # TODO: Maybe can be using other method that generally
    # delete files (?) Please, do if possible
    """
    if not filename or not is_file(filename):
        return None
    
    from os import remove as os_remove

    os_remove(filename)


from yta_general_utils.image.parser import ImageParser
from yta_multimedia.image.edition.resize import resize_image
from yta_multimedia.video.edition.resize import resize_video
from yta_multimedia.video.parser import VideoParser
from PIL import Image
from typing import Union

import numpy as np


class PixelFilterFunction:
    """
    Class to interact with image pixels and detect greens or transparent
    pixels to be used in, for example, ImageRegionFinder functionality.
    """
    @staticmethod
    def is_green(pixel):
        """
        This filter is the one we use to make sure it is a greenscreen
        color part by applying a [[0, r, 100], [100, g, 255], [0, b, 100]]
        filtering.
        """
        r, g, b = pixel

        return (r >= 0 and r <= 100) and (g >= 100 and g <= 255) and (b >= 0 and b <= 100)
    
    @staticmethod
    def is_transparent(pixel):
        """
        Checks if the alpha channel (4th in array) is set to 0 (transparent).
        The pixel must be obtained from a RGBA image (so 4 dimentions
        available).
        """
        _, _, _, a = pixel

        return a == 0

class Coordinate:
    """
    Class to represent a coordinate point (x, y).
    """
    x: int = None
    y: int = None

    def __init__(self, x: int, y: int):
        self.x = x
        self.y = y

class Region:
    """
    Class to represent a region built by two coordinates, one in
    the top left corner and another one in the bottom right 
    corner.
    """
    top_left: Coordinate = None
    bottom_right: Coordinate = None
    _width: int = None
    _height: int = None

    def __init__(self, top_left_x: int, top_left_y: int, bottom_right_x: int, bottom_right_y: int):
        self.top_left = Coordinate(top_left_x, top_left_y)
        self.bottom_right = Coordinate(bottom_right_x, bottom_right_y)
        self._width = self.bottom_right.x - self.top_left.x
        self._height = self.bottom_right.y - self.top_left.y

    @property
    def width(self):
        return self._width
    
    @property
    def height(self):
        return self._height
    
    def resize_image_to_fit_in(self, image):
        """
        This method rescales the provided 'image' to make it fit in
        this region. Once it's been rescaled, this image should be
        placed in the center of the region.
        """
        image = ImageParser.to_pillow(image)

        image = resize_image(image, (self.width, self.height))

        # We enlarge it by a 1% to avoid some balck pixels lines
        image = image.resize((image.size[0] * 1.01, image.size[1] * 1.01))

        return image

    # TODO: This could be private maybe
    def resize_video_to_fit_in(self, video):
        """
        This method rescales the provided 'video' to make it fit in
        this region. Once it's been rescaled, this video should be
        placed in the center of the region.
        """
        video = VideoParser.to_moviepy(video)

        video = resize_video(video, (self.width, self.height))

        # We enlarge it by a 1% to avoid some black pixels lines
        video = video.resize(1.01)

        return video
    
    def place_video_inside(self, video):
        """
        This method rescales the provided 'video' to make it fit in
        this region. Once it's been rescaled, this videos is 
        positioned in the required position to fit the region.
        """
        video = self.resize_video_to_fit_in(video)

        x = (self.bottom_right.x + self.top_left.x) / 2 - video.w / 2
        y = (self.bottom_right.y + self.top_left.y) / 2 - video.h / 2

        # TODO: What about upper limits (out of bottom left bounds) (?)
        if x < 0:
            x = 0
        if y < 0:
            y = 0

        video = video.set_position((x, y))

        return video

class ImageRegionFinder:
    directions = [(-1, 0), (1, 0), (0, -1), (0, 1), (-1, -1), (-1, 1), (1, -1), (1, 1)]

    @classmethod
    def is_valid(cls, x, y, image, visited, filter_func: callable):
        """
        This method verifies if the pixel is between the limits
        and is transparent and unvisited.
        """
        rows, cols, _ = image.shape

        return (0 <= x < rows and 0 <= y < cols and not visited[x, y] and filter_func(image[x, y]))

    @classmethod
    def dfs(cls, image: np.ndarray, visited, x, y, region, filter_func: callable):
        """
        A Deep First Search algorithm applied to the image to 
        obtain all the pixels connected in a region.
        """
        if not isinstance(image, np.ndarray):
            raise Exception('The provided "image" parameter is not a valid np.ndarray.')

        stack = [(x, y)]
        visited[x, y] = True
        region.append((x, y))
        
        while stack:
            cx, cy = stack.pop()
            for dx, dy in cls.directions:
                nx, ny = cx + dx, cy + dy
                if cls.is_valid(nx, ny, image, visited, filter_func):
                    visited[nx, ny] = True
                    region.append((nx, ny))
                    stack.append((nx, ny))

    @classmethod
    def is_inside(cls, small_bounds, large_bounds):
        """
        This method verifies if the bounds of a found region are
        inside another bounds to discard the smaller regions.
        """
        min_x_small, max_x_small, min_y_small, max_y_small = small_bounds
        min_x_large, max_x_large, min_y_large, max_y_large = large_bounds
        
        return (
            min_x_small >= min_x_large and max_x_small <= max_x_large and
            min_y_small >= min_y_large and max_y_small <= max_y_large
        )

    @classmethod
    def find_regions(cls, image: np.ndarray, filter_func: PixelFilterFunction) -> list[Region]:
        """
        This method looks for all the existing regions of transparent
        pixels that are connected ones to the others (neighbours). The
        'filter_func' parameter is the one that will classify the pixels
        as, for example, transparent or green. That 'filter_func' must
        be a method contained in the PixelFilterFunction class.

        This method returns the found regions as objects with 'top_left'
        and 'bottom_right' fields that are arrays of [x, y] positions
        corresponding to the corners of the found regions.
        """
        if not isinstance(image, np.ndarray):
            raise Exception('The provided "image" parameter is not a valid np.ndarray.')

        rows, cols, _ = image.shape
        visited = np.zeros((rows, cols), dtype=bool)
        regions = []
        
        for row in range(rows):
            for col in range(cols):
                # If we find a transparent pixel, we search
                if filter_func(image[row, col]) and not visited[row, col]:
                    region = []
                    cls.dfs(image, visited, row, col, region, filter_func)
                    
                    if region:
                        min_x = min(px[0] for px in region)
                        max_x = max(px[0] for px in region)
                        min_y = min(px[1] for px in region)
                        max_y = max(px[1] for px in region)
                        
                        # These are the limits of the region
                        bounds = (min_x, max_x, min_y, max_y)
                        
                        # We need to avoid small regions contained in others
                        if not any(cls.is_inside(bounds, r['bounds']) for r in regions):
                            regions.append({
                                # TODO: Maybe we need them to turn into transparent pixels
                                #'coordinates': region, # We don't need coordinates
                                'bounds': bounds
                            })

        # I want another format, so:
        for index, region in enumerate(regions):
            regions[index] = Region(region['bounds'][2], region['bounds'][0], region['bounds'][3], region['bounds'][1])
            # regions[index] = {
            #     # 'top_left': [region['bounds'][0], region['bounds'][2]],
            #     # 'bottom_right': [region['bounds'][1], region['bounds'][3]]
            #     # I don't know why I have to use it in this order but...
            #     'top_left': [region['bounds'][2], region['bounds'][0]],
            #     'bottom_right': [region['bounds'][3], region['bounds'][1]]
            # }

        return regions
    
    @classmethod
    def find_green_regions(cls, image: Union[str, Image.Image, np.ndarray]) -> list[Region]:
        """
        This method returns the found green regions as objects with
        'top_left' and 'bottom_right' fields that are arrays of [x, y] 
        positions corresponding to the corners of the found regions.
        """
        image = ImageParser.to_numpy(image, 'RGB')
            
        return cls.find_regions(image, PixelFilterFunction.is_green)
    
    @classmethod
    def find_transparent_regions(cls, image: Union[str, Image.Image, np.ndarray]) -> list[Region]:
        """
        This method returns the found transparent regions as objects
        with 'top_left' and 'bottom_right' fields that are arrays of
        [x, y] positions corresponding to the corners of the found
        regions.
        """
        image = ImageParser.to_numpy(image, 'RGBA')
            
        return cls.find_regions(image, PixelFilterFunction.is_transparent)


from yta_general_utils.web.scrapper.chrome_scrapper import google_translate
from enum import Enum
from typing import Union


class GoogleLanguage(Enum):
    # TODO: Implement more languages
    ENGLISH = 'en'
    SPANISH = 'es'

def translate_text(text: str, input_language: Union[GoogleLanguage, str] = GoogleLanguage.ENGLISH, output_language: Union[GoogleLanguage, str] = GoogleLanguage.SPANISH):
    """
    Returns the provided 'text' translated into the 'output_language' 
    using Google Traductor by chromedriver navigation.

    TODO: Make this have Enums with the different available Google
    Translate languages.
    """
    if not text:
        raise Exception('No "text" provided.')
    
    if not isinstance(text, str) and not isinstance(text, GoogleLanguage):
        raise Exception('The "text" provided is not a GoogleLanguage or a str.')
    
    if not input_language:
        raise Exception('No "input_language" provided.')
    
    if not output_language:
        raise Exception('No "output_language" provided.')

    return google_translate(text, input_language, output_language)
